import logging

logger = logging.getLogger(__name__)


class UncallableHook(Exception):
    pass


class BaseHook(object):
    hook_type = None
    callable = None
    _key = None

    def __init__(self, callable=None):
        if callable:
            self.callable = callable

    def run(self, hints, cache, **kwargs):
        if self.callable:
            try:
                if 'key' not in kwargs:
                    kwargs['key'] = self.key()
                self.callable(hints=hints, cache=cache, **kwargs)
            except TypeError:
                raise UncallableHook()

    def run_if_type_matches(self, hook_type, hints, cache, **kwargs):
        if hook_type != self.hook_type:
            return
        self.run(hints=hints, cache=cache, **kwargs)

    def key(self, k=None):
        if k:
            self._key = k
        return self._key


class PreSetHook(BaseHook):
    hook_type = 'pre-set'


class PostSetHook(BaseHook):
    hook_type = 'post-set'


class PrePrimeHook(BaseHook):
    hook_type = 'pre-prime'


class PostPrimeHook(BaseHook):
    hook_type = 'post-prime'


class PostSetRefreshHook(PostSetHook):
    def __init__(self, refresh_key=None, param_func=None):
        self.refresh_key = refresh_key
        self.param_func = param_func

    def run(self, hints, cache, **kwargs):
        if self.refresh_key:
            key = cache.build_key(
                self.refresh_key,
                **self.param_func(**cache.primer_instance(self.key()).parameter_dict()))
            logger.debug('PostSetRefreshHook.run() - refreshing key: {}'.format(key))
            cache.refresh(key, queued=True)



# Generated by Django 2.2.5 on 2019-09-27 05:15

from django.db import migrations
import picklefield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('django_kck', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dataproduct',
            name='value',
            field=picklefield.fields.PickledObjectField(editable=False),
        ),
        migrations.AlterField(
            model_name='refreshqueue',
            name='hints',
            field=picklefield.fields.PickledObjectField(blank=True, editable=False, null=True),
        ),
    ]



import datetime
import logging
import dateutil

logger = logging.getLogger(__name__)


class HintsManager(object):
    """
    The HintsManager addresses the issue that has come up where the concept of a 'hint'
    is used to suggest some bit of information that is being propagated to avoid a
    relatively-expensive query, but the implementation varies enough to where it's
    getting tricky to keep track of exactly what a hint means.

    This class nails down the concept of a hint to mean exactly a substitution of a result
    for a specific key that would otherwise have been resolved via a cache hit.

    ***Further, the sets record the timestamp and the get method can restrict its use of the
    hints by time range.***

    The hint can handle both sets and deletes.
    """

    _hints_by_key = {}
    _hints_by_timestamp = {}
    _purged = {}
    _cache_obj = None

    def __init__(self, cache_obj):
        self._cache_obj = cache_obj

    def import_from_dict(self, hints_dict):

        # import hints
        for key, cache_records in hints_dict['hints'].items():
            for cache_record in cache_records:
                if key not in self._hints_by_key:
                    self.set_hint_key(cache_record)
                    continue

        # import purged
        for key, datetime_list in hints_dict['purged'].items():
            self.unset_hint_key(key, max(datetime_list))

    def set_hint_key(self, cache_entry=None):
        """setter for values"""

        if cache_entry['key'] not in self._hints_by_key:
            self._hints_by_key[cache_entry['key']] = [cache_entry]
        else:
            self._hints_by_key[cache_entry['key']].append(cache_entry)

        # self._hints_by_key[cache_entry['key']] = (cache_entry, cache_entry['modified'])

        if cache_entry['modified'] not in self._hints_by_timestamp:
            self._hints_by_timestamp[cache_entry['modified']] = []
        self._hints_by_timestamp[cache_entry['modified']].append(cache_entry)

    def unset_hint_key(self, key=None, unset_datetime=None):
        """setter for deletions"""
        unset_datetime = unset_datetime or datetime.datetime.utcnow()
        if key not in self._purged:
            self._purged[key] = [unset_datetime]
        else:
            self._purged[key].append(unset_datetime)

    def count(self, since=None, exclude_keys=None):
        """ count number of hints with or without since param """
        ret = 0
        for ts in self._hints_by_timestamp.keys():
            if not since or ts >= since:
                ret += len(self._hints_by_timestamp[ts])

        if exclude_keys:
            for key in exclude_keys:
                if key in self._hints_by_key:
                    ret -= 1

        return ret

    def hinted_keys(self):
        return self._hints_by_key.keys()

    def get(self, key, prime_on_cache_miss=False, hints_only=False):
        last_purge = None
        if key in self._purged:
            last_purge = max(self._purged[key])
        max_modified_date = None
        most_recent_cache_record = None

        if key not in self._hints_by_key:

            if hints_only:
                raise KeyError(key)

            cache_entry = self._cache_obj.get(key, prime_on_cache_miss=prime_on_cache_miss)
            if ('modified' in cache_entry and cache_entry['modified'] and
                    last_purge and last_purge > cache_entry['modified']):
                raise KeyError()
            self.set_hint_key(cache_entry)
            return cache_entry

        for cache_record in self.get_all_hints(key):
            logger.debug('cache record for key: {} - {}'.format(key, cache_record))

            if type(cache_record['modified']) is str:
                cache_record['modified'] = dateutil.parser.parse(cache_record['modified'])
            if last_purge and cache_record['modified'] < last_purge:
                continue
            if max_modified_date is None or max_modified_date < cache_record['modified']:
                max_modified_date = cache_record['modified']
                most_recent_cache_record = cache_record

        if most_recent_cache_record:
            return most_recent_cache_record

        return self._cache_obj.get(key, prime_on_cache_miss=prime_on_cache_miss)

    def get_all_hints(self, key):
        return self._hints_by_key[key]

    def find_hint(self, key, match_entry, sort_key=None, index=0, invert=False,
                  value_filters=None, disregard_keys=None):
        try:
            unsorted_cache_entries = self.get_all_hints(key)
        except KeyError:
            return None
        if not unsorted_cache_entries:
            return None
        if invert:
            cache_entries = sorted(
                filter(
                    lambda e: not self._cache_obj.cmp_cache_entries(
                        e,
                        match_entry,
                        value_filters=value_filters,
                        disregard_keys=disregard_keys),
                    unsorted_cache_entries),
                key=sort_key)
        else:
            cache_entries = sorted(
                filter(
                    lambda e: self._cache_obj.cmp_cache_entries(
                        e,
                        match_entry,
                        value_filters=value_filters,
                        disregard_keys=disregard_keys),
                    unsorted_cache_entries),
                key=sort_key)
        try:
            return cache_entries[index]
        except IndexError:
            return None

    def to_dict(self):
        """dump to dict"""
        return {'hints': self._hints_by_key, 'purged': self._purged}

    def from_dict(self, d):
        """load dict"""
        if 'hints' in d:
            self._hints_by_key = d['hints']
        if 'purged' in d:
            self._purged = d['purged']

    def update_from_dict(self, d):
        """add keys from dict d to internal dict"""
        if 'hints' in d:
            self._hints_by_key = {**self._hints_by_key, **d['hints']}
        if 'purged' in d:
            self._purged = {**self._purged, **d['purged']}



from django.db import models
from picklefield.fields import PickledObjectField


class DataProduct(models.Model):
    """the cache table"""
    key = models.CharField(max_length=1024, primary_key=True)
    value = PickledObjectField()
    primer_name = models.CharField(max_length=1024, db_index=True, null=True, blank=True)
    version = models.PositiveIntegerField(default=1)
    modified = models.DateTimeField(auto_now_add=True)
    soft_expire = models.DateTimeField(null=True, blank=True)
    hard_expire = models.DateTimeField(null=True, blank=True)

    class Meta:
        unique_together = ('primer_name', 'key')


class RefreshQueue(models.Model):
    key = models.CharField(max_length=1024)
    hints = PickledObjectField(blank=True, null=True)
    primer_name = models.CharField(max_length=1024, db_index=True)
    claimant = models.CharField(max_length=64, null=True, blank=True)
    claimed = models.DateTimeField(null=True, blank=True)
    requested = models.DateTimeField(auto_now_add=True)



# Generated by Django 2.2.5 on 2019-09-24 03:40

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RefreshQueue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=1024)),
                ('hints', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('primer_name', models.CharField(db_index=True, max_length=1024)),
                ('claimant', models.CharField(blank=True, max_length=64, null=True)),
                ('claimed', models.DateTimeField(blank=True, null=True)),
                ('requested', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='DataProduct',
            fields=[
                ('key', models.CharField(max_length=1024, primary_key=True, serialize=False)),
                ('value', django.contrib.postgres.fields.jsonb.JSONField()),
                ('primer_name', models.CharField(blank=True, db_index=True, max_length=1024, null=True)),
                ('version', models.PositiveIntegerField(default=1)),
                ('modified', models.DateTimeField(auto_now_add=True)),
                ('soft_expire', models.DateTimeField(blank=True, null=True)),
                ('hard_expire', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'unique_together': {('primer_name', 'key')},
            },
        ),
    ]
